import os
from argparse import Namespace
from typing import List, Optional

import yaml

class DataConfig:
    def __init__(
        self,
        surface_vars = [],
        static_surface_vars = [],
        vertical_vars = [],
        levels = [],
        time_range_train = [],
        time_range_valid = [],
        **kwargs,
    ):
        self.__dict__.update(kwargs)

        self.surface_vars = surface_vars
        self.static_surface_vars = static_surface_vars
        self.vertical_vars = vertical_vars
        self.levels = levels
        self.time_range_train = time_range_train
        self.time_range_valid = time_range_valid

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def from_argparse(args: Namespace):
        return DataConfig(**args.__dict__)


class ModelConfig:
    def __init__(
        self,
        num_static_channels: Optional[int] = None,
        embed_dim: Optional[int] = None,
        token_size: Optional[tuple[int, int]] = None,
        n_blocks_encoder: Optional[int] = None,
        n_blocks_decoder: Optional[int] = None,
        mlp_multiplier: Optional[int] = None,
        n_heads: Optional[int] = None,
        dropout_rate: Optional[float] = None,
        residual: Optional[bool] = False,
        train_loss: Optional[str] = None,
        val_loss: Optional[str] = None,
        **kwargs,
    ):
        self.__dict__.update(kwargs)

        self.num_static_channels = num_static_channels
        self.embed_dim = embed_dim
        self.token_size = token_size
        self.n_blocks_encoder = n_blocks_encoder
        self.n_blocks_decoder = n_blocks_decoder
        self.mlp_multiplier = mlp_multiplier
        self.n_heads = n_heads
        self.dropout_rate = dropout_rate
        self.residual = residual
        self.train_loss = train_loss
        self.val_loss = val_loss

        self.__dict__.update(kwargs)

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def from_argparse(args: Namespace):
        return ModelConfig(**args.__dict__)

    @property
    def encoder_d_ff(self):
        return int(self.enc_embed_size * self.mlp_ratio)

    @property
    def decoder_d_ff(self):
        return int(self.dec_embed_size * self.mlp_ratio)

    def __str__(self):
        return (
            f"Input channels: {self.num_input_channels}, "
            f"Encoder (L, H, E): {[self.enc_num_layers, self.enc_num_heads, self.enc_embed_size]}, "
            f"Decoder (L, H, E): {[self.dec_num_layers, self.dec_num_heads, self.dec_embed_size]}"
        )

    def __repr__(self):
        return (
            f"Input channels: {self.num_input_channels}, "
            f"Encoder (L, H, E): {[self.enc_num_layers, self.enc_num_heads, self.enc_embed_size]}, "
            f"Decoder (L, H, E): {[self.dec_num_layers, self.dec_num_heads, self.dec_embed_size]}"
        )


class ExperimentConfig:
    def __init__(
        self,
        job_id: str = "",
        data_config: DataConfig = None,
        model_config: ModelConfig = None,
        num_epochs: int = 1,
        limit_steps_train: int = 1,
        limit_steps_valid: int = 1,
        batch_size: int = 1,
        learning_rate: float = 1e-3,
        min_lr: float = 1e-5,
        dl_num_workers: int = 1,
        dl_prefetch_size: int = 0,
        path_experiment: str = "./",
        warm_up_steps: int = 0,
        mask_unit_size: Optional[tuple[int]] = None,
        mask_ratio_inputs: Optional[float] = None,
        mask_ratio_targets: Optional[float] = None,
        **kwargs,
    ):
        # additional experiment parameters used in downstream tasks
        self.__dict__.update(kwargs)

        self.job_id = job_id
        self.data = data_config
        self.model = model_config
        self.num_epochs = num_epochs
        self.limit_steps_train = limit_steps_train
        self.limit_steps_valid = limit_steps_valid
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        self.min_lr = min_lr
        self.warm_up_steps = warm_up_steps
        self.dl_num_workers = dl_num_workers
        self.dl_prefetch_size = dl_prefetch_size
        self.mask_unit_size = mask_unit_size
        self.mask_ratio_inputs = mask_ratio_inputs
        self.mask_ratio_targets = mask_ratio_targets
        self.path_experiment = path_experiment

    @property
    def path_checkpoint(self) -> str:
        if self.path_experiment == '':
            return os.path.join(self.path_weights, 'train', 'checkpoint.pt')
        else:
            return os.path.join(
                os.path.dirname(self.path_experiment), 'weights', 'train', 'checkpoint.pt'
            )

    @property
    def path_weights(self) -> str:
        return os.path.join(self.path_experiment, self.make_suffix_path(), "weights")

    @property
    def path_wandb(self) -> str:
        return os.path.join(self.path_experiment, self.make_suffix_path())

    def to_dict(self):
        d = self.__dict__.copy()
        d["model"] = self.model.to_dict()
        d["data"] = self.data.to_dict()

        return d

    @staticmethod
    def from_argparse(args: Namespace):
        return ExperimentConfig(
            data_config=DataConfig.from_argparse(args),
            model_config=ModelConfig.from_argparse(args),
            **args.__dict__,
        )

    @staticmethod
    def from_dict(params: dict):
        return ExperimentConfig(
            data_config=DataConfig(**params['data']),
            model_config=ModelConfig(**params['model']),
            **params,
        )

    def make_folder_name(self) -> str:
        param_folder = f"v1"
        return param_folder

    def make_suffix_path(self) -> str:
        return os.path.join(self.make_folder_name(), self.job_id)

    def __str__(self):
        return (
            f"ID: {self.job_id}, "
            f"Epochs: {self.num_epochs}, "
            f"Truncate train: {self.limit_steps_train}, "
            f"Truncate valid: {self.limit_steps_valid}, "
            f"Batch size: {self.batch_size}, "
            f"LR: {self.learning_rate}, "
            f"DL workers: {self.dl_num_workers}"
        )

    def __repr__(self):
        return (
            f"ID: {self.job_id}, "
            f"Epochs: {self.num_epochs}, "
            f"Truncate train: {self.limit_steps_train}, "
            f"Truncate valid: {self.limit_steps_valid}, "
            f"Batch size: {self.batch_size}, "
            f"LR: {self.learning_rate}, "
            f"DL workers: {self.dl_num_workers}"
        )


def get_config(config_path: str) -> ExperimentConfig:
    cfg = yaml.safe_load(open(config_path, 'r'))
    return ExperimentConfig.from_dict(cfg)



"""granite-wxc - Weather and climate downscaling model."""

__version__ = "0.1.0"

from . import datasets, models, decoders, utils

__all__ = [
    "datasets",
    "models",
    "decoders",
    "utils"
]


import os
import random
from datetime import timedelta
from functools import partial

import numpy as np
import torch
import torch.backends.cudnn as cudnn
import torch.distributed as dist
from torch.distributed.fsdp import FullyShardedDataParallel, ShardingStrategy, MixedPrecision
from torch.distributed.fsdp.wrap import size_based_auto_wrap_policy


class _MPI:
    def __getattr__(self, attr):
        from mpi4py import MPI

        return getattr(MPI, attr)


mpi = _MPI()


def init_dist(device: str, rank: int, world_size: int):
    torch.distributed.init_process_group(
        backend=device,
        init_method='env://',
        world_size=world_size,
        rank=rank,
        timeout=timedelta(seconds=600),
    )


def _get_local_rank():
    envs = [
        'MV2_COMM_WORLD_LOCAL_RANK',
        'OMPI_COMM_WORLD_LOCAL_RANK',
        'SLURM_LOCALID',
        'LOCAL_RANK',
    ]

    for ev in envs:
        if ev in os.environ:
            return int(os.environ[ev])
    else:
        from mpi4py import MPI
        return MPI.COMM_WORLD.Split_type(MPI.COMM_TYPE_SHARED).rank


def _get_rank():
    try:
        return int(os.environ["RANK"])
    except KeyError:
        from mpi4py import MPI
        comm = MPI.COMM_WORLD
        return comm.rank


def _get_world_size():
    try:
        return int(os.environ["WORLD_SIZE"])
    except KeyError:
        from mpi4py import MPI
        return MPI.COMM_WORLD.size


def init_ddp(use_gpu: bool):
    local_rank = _get_local_rank()
    rank = _get_rank()
    world_size = _get_world_size()

    if use_gpu:
        assert torch.cuda.is_available(), "GPU requested but none was found in the system."

    if use_gpu:
        init_dist("nccl", rank, world_size)
        torch.cuda.set_device(local_rank)
        os.environ["TORCH_SHOW_CPP_STACKTRACES"] = str(1)
        os.environ["TORCH_NCCL_ASYNC_ERROR_HANDLING"] = str(1)
        os.environ["TORCH_DISTRIBUTED_DEBUG"] = "DETAIL"
        cudnn.benchmark = True
    else:
        init_dist("gloo", rank, world_size)
    return local_rank, rank


def set_global_seed(rank):
    random.seed(42 + rank)
    torch.cuda.manual_seed(42 + rank)
    torch.manual_seed(42 + rank)
    np.random.seed(42 + rank)


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def get_world_size():
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()


def get_local_world_size():
    local_world_size = os.environ.get('LOCAL_WORLD_SIZE')
    if local_world_size is None:
        return 1
    else:
        return int(local_world_size)


def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()


def get_node_rank():
    node_rank = os.environ.get('GROUP_RANK')
    if node_rank is None:
        return 0
    else:
        return int(node_rank)


def get_local_rank():
    local_rank = os.environ.get('LOCAL_RANK')
    if local_rank is None:
        return 0
    else:
        return int(local_rank)


def is_main_process():
    return get_rank() == 0


def model_distribute(
    model: torch.nn.Module, use_gpu: bool=False
):

    if not use_gpu:
        raise RuntimeError("For FSDP currently we require GPU support")

    wrap_policy = partial(
        size_based_auto_wrap_policy, min_num_params=10_000,
    )
    fsdp_kwargs_encoder_decoder = {
        # Options:
        # FULL_SHARD, SHARD_GRAD_OP, NO_SHARD, HYBRID_SHARD, _HYBRID_SHARD_ZERO2
        "sharding_strategy": ShardingStrategy.FULL_SHARD,
        "device_id": torch.cuda.current_device(),
        "use_orig_params": False,
        "auto_wrap_policy": wrap_policy,
        "mixed_precision" : MixedPrecision(
            param_dtype=torch.bfloat16,
            #reduce_dtype=torch.bfloat16,
            #buffer_dtype=torch.bfloat16,
            cast_forward_inputs=True
        )
    }
    fsdp_kwargs_unembed = {
        # Options:
        # FULL_SHARD, SHARD_GRAD_OP, NO_SHARD, HYBRID_SHARD, _HYBRID_SHARD_ZERO2
        "sharding_strategy": ShardingStrategy.FULL_SHARD,
        "device_id": torch.cuda.current_device(),
        "use_orig_params": False,
        #"auto_wrap_policy": wrap_policy,
        "mixed_precision" : MixedPrecision(
            param_dtype=torch.float32,
            #reduce_dtype=torch.float32,
            #buffer_dtype=torch.float32,
            cast_forward_inputs=True
        )
    }
    fsdp_kwargs_model = {
        # Options:
        # FULL_SHARD, SHARD_GRAD_OP, NO_SHARD, HYBRID_SHARD, _HYBRID_SHARD_ZERO2
        "sharding_strategy": ShardingStrategy.FULL_SHARD,
        "device_id": torch.cuda.current_device(),
        "use_orig_params": False,
        #"auto_wrap_policy": wrap_policy,
        "mixed_precision" : MixedPrecision(
            param_dtype=torch.float32,
            #reduce_dtype=torch.float32,
            #buffer_dtype=torch.float32,
            cast_forward_inputs=False
        )
    }

    device = torch.cuda.current_device()
    torch.cuda.empty_cache()
    model = model.to(device)
    model.encoder = FullyShardedDataParallel(model.encoder, **fsdp_kwargs_encoder_decoder)
    model.decoder = FullyShardedDataParallel(model.decoder, **fsdp_kwargs_encoder_decoder)
    model.unembed = FullyShardedDataParallel(model.unembed, **fsdp_kwargs_unembed)
    model = FullyShardedDataParallel(model, **fsdp_kwargs_model)

    return model


from typing import List, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

UPSAMPLING_MODES = ['nearest', 'bilinear', 'pixel_shuffle', 'conv_transpose']

class PixelShuffleBlock(nn.Module):
    def __init__(self, scale: int, in_channels: int, channels: int, kernel_size: int = 3, stride: int = 1,
                 activation: Optional[nn.Module] = nn.PReLU):
        """ Upsampling with PixelShuffle: conv -> shuffle -> activation

        Args:
            scale: (int) upscale factor.
            in_channels: (int) input channels.
            channels: (int) number of output channels.
            kernel_size: (int) kernel size for the convolution op.
            stride: (int) stride for the convolution op.
            activation: (int) activation after the convolution op (optional).
        """
        super().__init__()

        self.conv = nn.Conv2d(in_channels, channels * scale ** 2,
                              kernel_size=kernel_size, stride=stride, padding='same', padding_mode='replicate')
        self.pixel_shuffle = nn.PixelShuffle(scale)
        self.activation = activation() if activation is not None else nn.Identity()

    def forward(self, x):
        x = self.conv(x)
        x = self.pixel_shuffle(x)
        x = self.activation(x)

        return x


class InterpBlock(nn.Module):
    def __init__(self, scale: int,  in_channels: int, channels: int, kernel_size: int = 3, stride: int = 1,
                 activation: Optional[nn.Module] = nn.ReLU, mode='bilinear'):
        """ Upsampling with interpolation: interpolation -> conv -> activation

        Args:
            scale: (int) upscale factor.
            in_channels: (int) input channels.
            channels: (int) number of output channels for the convolution op.
            kernel_size: (int) kernel size for the convolution op.
            stride: (int) stride for the convolution op.
            activation: (int) activation after the convolution op (optional).
        """
        super().__init__()

        self.conv = nn.Conv2d(in_channels, channels, kernel_size=kernel_size, stride=stride, padding='same', padding_mode='replicate')
        self.upsample = nn.Upsample(scale_factor=scale, mode=mode)
        self.activation = activation() if activation is not None else nn.Identity()

    def forward(self, x):
        x = self.upsample(x)
        x = self.conv(x)
        x = self.activation(x)

        return x


class ConvTransposeBlock(nn.Module):
    def __init__(self, in_channels: int, channels: int, kernel_size: int = 2, stride: int = 2,
                 activation: Optional[nn.Module] = nn.ReLU):
        """ Upsampling with ConvTranspose2d: conv -> activation
            (stride must be > 1 to increase spatial size)

        Args:
            in_channels: (int) input channels.
            channels: (int) number of output channels for the convolution op.
            kernel_size: (int) kernel size for the convolution op.
            stride: (int) stride for the convolution op.
            activation: (int) activation after the convolution op (optional).
        """
        super().__init__()

        self.conv = nn.ConvTranspose2d(in_channels, channels, kernel_size=kernel_size, stride=stride)
        self.activation = activation() if activation is not None else nn.Identity()

    def forward(self, x: torch.Tensor):
        x = self.conv(x)
        x = self.activation(x)

        return x


class ConvEncoderDecoder(nn.Module):
    def __init__(
            self,
            in_channels: int,
            channels: int,
            out_channels: int,
            scale: List[int],
            upsampling_mode: str,
            kernel_size: List[int],
            stride: int = 1,
            activation: Optional[nn.Module] = nn.PReLU,
        ):
        """ Simple convolutional decoder for downscaling.

        Args:
            scale: list with upscale factors for each stage.
            in_channels: (int) input channels.
            channels: (int) number of channels for the convolution ops.
            upsampling_mode: (str) upsampling to use ('nearest', 'bilinear', 'pixel_shuffle', 'conv_transpose').
            kernel_size: (int) kernel size for the convolution ops.
            stride: (int) stride for the convolution ops.
            activation: (int) activation after each convolution ops (optional).
            out_channels: (int) number of output channels (target vars). Default = 1.
        """

        super().__init__()

        self.scale = scale
        self.kernel_size = kernel_size 
        if upsampling_mode not in UPSAMPLING_MODES:
            raise ValueError(f"Incorrect upsampling mode {upsampling_mode}.")
        self.upsampling_mode = upsampling_mode
        
        # build layers
        current_ch = in_channels
        self.blocks = nn.ModuleList()
        for k_i, s_i in zip(self.kernel_size, self.scale):
            if upsampling_mode == 'pixel_shuffle':
                layer = PixelShuffleBlock(scale=s_i, in_channels=current_ch, channels=channels,
                                          kernel_size=k_i, stride=stride, activation=activation)
            elif upsampling_mode == 'conv_transpose':
                layer = ConvTransposeBlock(in_channels=current_ch, channels=channels,
                                           kernel_size=k_i, stride=s_i, activation=activation)
            else:
                layer = InterpBlock(scale=s_i, in_channels=current_ch, channels=channels, kernel_size=k_i,
                                    stride=stride, activation=activation, mode=upsampling_mode)
            current_ch = channels

            self.blocks.append(layer)

        # Assuming downscaling of 1 variable
        self.out_conv = nn.Conv2d(in_channels=current_ch, out_channels=out_channels,
                                  kernel_size=3, stride=1, padding='same', padding_mode='replicate')

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Conv2d) or isinstance(m, nn.Linear):
            nn.init.trunc_normal_(m.weight, std=.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        '''
        Args:
            x: Input of shape (batch, in_channels, lat, lon)
        Returns:
            Tensor of shape (batch, out_channels, scale*lat, scale*lon)
        '''

        for block in self.blocks:
            x = block(x)
        x = self.out_conv(x)

        return x


def token_to_pixel(x: torch.Tensor, patch_size_px: List[int], batch_size: int,
                   global_shape_mu: List[int], local_shape_mu: List[int]):
    """ Tokens in global-local sequence to pixel space.
        (B, global seq, local seq, C * patch_sizes) -> (B, C, H, W)

    Args:
        x: torch tensor of shape (B, global seq, local seq, emb_dim * patch_sizes)
        patch_size_px: list with patch size in each dimension (in pixels).
        batch_size: (int) batch size
        global_shape_mu: list with mask unit shape.
        local_shape_mu: list with tokens shape inside a mask unit.

    Returns:
        torch tensor in pixel space; shape = (B, C, H, W)
    """
    # Reshape to (batch, global_lat, global_lon, local_lat, local_lon, in_channels * patch_size[0] * patch_size[1])
    assert x.shape[0] == batch_size
    assert x.shape[1] == global_shape_mu[0] * global_shape_mu[1]
    assert x.shape[2] == local_shape_mu[0] * local_shape_mu[1]
    x = x.reshape(
        batch_size,
        global_shape_mu[0],
        global_shape_mu[1],
        local_shape_mu[0],
        local_shape_mu[1],
        -1
    )
    
    # (batch, global seq, local seq, in_channel) -> (batch, channel, lat, lon)
    # Permute to (batch, in_channels * patch_size[0] * patch_size[1], global_lat, local_lat, global_lon, local_lon)
    x = x.permute(0, 5, 1, 3, 2, 4)
    # Reshape to (batch, in_channels * patch_size[0] * patch_size[1], lat // patch_size[0], lon // patch_size[1])
    x = x.flatten(4, 5).flatten(2, 3)
    # Pixel shuffle to (batch, in_channels, lat, lon)
    if patch_size_px[0] != patch_size_px[1]:
        raise NotImplementedError('Current pixel shuffle implementation assumes same patch size along both dimensions.')
    x = F.pixel_shuffle(x, patch_size_px[0])

    return x



import time
import numpy as np
import scipy.stats as stats

import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.animation as animation

def plot_spatial(sample, ax, title, **kwargs):
    
    vmin=kwargs.get('vmin', np.min(sample))
    vmax=kwargs.get('vmax', np.max(sample))
    
    if kwargs.get('c_lognorm', False):
        vmin = max(vmin, 0)
        e = kwargs.get('e', 1e-44)
        vmin+=e
        norm=colors.LogNorm(vmin=vmin, vmax=vmax)
    else:
        norm=colors.Normalize(vmin=vmin, vmax=vmax)
        
    im = ax.imshow(
        sample,
        cmap=kwargs.get('cmap', 'coolwarm'),
        norm=norm,
        origin='lower',
        animated=kwargs.get('animated', False),
    )
    ax.set_xlabel('Longitudes')
    ax.set_ylabel('Latitudes')
    ax.tick_params(axis='both', which='major', labelsize=10)
    ax.title.set_text(title)
    
    return im

def plot_model_residual(sample, sample_id, title, **kwargs):

    fig, ax = plt.subplots(figsize=(15, 5))
    plt.suptitle(title)
    
    im = plot_spatial(sample, ax, sample_id,  **kwargs)
    
    label = f"{kwargs.get('var_name_tile', '')} [{kwargs.get('var_unit', '')}]"
    fig.colorbar(im, ax=ax, orientation='vertical', label=label, fraction=0.05, aspect=50)
    
    

def plot_model_results(samples, samples_id, title, **kwargs):
    
    fig, axes = plt.subplots(1, len(samples), figsize=(20, 5))
    plt.suptitle(title)
    
    for i, ax in enumerate(axes):
        if samples_id[i] == 'Residual':
            im_res = plot_spatial(samples[i], ax, samples_id[i],  **kwargs.get('plot_residual_kwargs'))
            label = f"{kwargs.get('var_name_tile', '')} [{kwargs.get('var_unit', '')}]"
            fig.colorbar(im_res, ax=ax, orientation='vertical', label=label, fraction=0.015)
        else:
            im = plot_spatial(samples[i], ax, samples_id[i],  **kwargs)
        
    label = f"{kwargs.get('var_name_tile', '')} [{kwargs.get('var_unit', '')}]"
    fig.colorbar(im, ax=axes, orientation='horizontal', label=label, fraction=0.05, aspect=50)
 
    plt.close()
    
    return fig

def plot_power_spectrum(img, ax, save_fig=False):
    """
    A power spctrum mesaures the strength of features at different resolutions

    :param img: H x W
    """

    npix = img.shape[-2], img.shape[-1]

    fft_img = np.fft.fftn(img)
    fft_amp = np.abs(fft_img)**2
    fft_amp = fft_amp.flatten()

    kfreq_x = np.fft.fftfreq(npix[1]) * npix[1] # wave vector
    kfreq_y = np.fft.fftfreq(npix[0]) * npix[0] # wave vector
    kfreq2D = np.meshgrid(kfreq_x, kfreq_y)
    knrm = np.sqrt(kfreq2D[0]**2 + kfreq2D[1]**2)
    knrm = knrm.flatten()

    kbins = np.arange(0.5, min(*npix)//2+1, 1.)
    kvals = 0.5 * (kbins[1:] + kbins[:-1])
    Abins, _, _ = stats.binned_statistic(
        knrm,
        fft_amp,
        statistic='mean',
        bins=kbins
    )

    Abins*= np.pi * (kbins[1:]**2 - kbins[:-1]**2)

    ax.loglog(kvals, Abins)
    ax.set_xlabel("$k$")
    ax.set_ylabel("$P(k)$")
    plt.tight_layout()

    if save_fig:
        timestr = time.strftime("%Y%m%d-%H%M")
        plt.savefig(f'power_spectrum_{timestr}.png', dpi=300, bbox_inches='tight')

def spatial_rmse(y_hat, y):
    return np.mean((y_hat - y) ** 2) ** 0.5
    
def spatial_bias(y_hat, y):
    return y_hat.mean() - y.mean()


import os
from typing import Optional

import torch
from torch import nn

from granitewxc.utils import distributed

class PatchEmbed(torch.nn.Module):
    '''
    Patch embedding via 2D convolution.
    same as weather-fm.models.hiera_maxvit.PatchEmbed
    '''

    def __init__(self, patch_size: int | tuple[int, ...], channels: int, embed_dim: int):
        super().__init__()

        self.patch_size = patch_size
        self.channels = channels
        self.embed_dim = embed_dim

        self.proj = nn.Conv2d(
            channels, embed_dim, kernel_size=patch_size, stride=1, bias=True, padding='same', padding_mode='replicate'
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        '''
        Args:
            x: Tensor of shape [batch, channels, lat, lon].
        Returns:
            Tensor with shape [batch, embed_dim, lat//patch_size, lon//patch_size]
        '''

        x = self.proj(x)

        return x


class FinetuneWrapper(torch.nn.Module):
    """ General purpose wrapper class to finetune using us configurable head and backbone """

    def __init__(self, backbone: torch.nn.Module, head: torch.nn.Module, config=None):
        super().__init__()

        self.backbone = backbone  # pre-trained model
        self.head = head  # task specific head
        self.config = config
        
        # weight_path = Path(config.load_model)
        # if os.path.isfile(weight_path):
        #     cp_path = weight_path
        # else:
        #     cp_path = weight_path / "train"
        # self.checkpointer = Checkpointer(
        #     cp_path, 2, "fsdp", 0, 0
        # )

    def forward(self, batch):
        """
        Args:
            batch: dependent on the backbone
        Returns:
            output of the configured head
        """
        raise NotImplementedError('Forward for wrapper has not been implemented.')
    
    
    def load_pretrained_backbone(
            self, 
    ):
        """  Based off of load checkpoint

        Args:
            weights_path: path to model checkpoint with only model weights
            ignore_modules: modules to ignore within the selected hierarchy.
                To ignore embedding related modules, set variable to ['patch_embedding', 'unembed']
            sel_prefix: '' selects all the modules within PrithviWxC. 'encoder.' selects encoder only.
            freeze: freezes the backbone when set.
            return_keys: If True, returns the output of load_state_dict(): missing_keys and unexpected_keys fields.

        Returns:
            If *return_keys* is True, `NamedTuple`` with ``missing_keys`` and ``unexpected_keys`` fields.
        """
        if distributed.is_main_process():
            print(f"Loading pre-trained model weights {self.config.load_model}...")

        if self.config.load_exp:
            self.backbone, _, _, _, last_epoch, _, _ = self.checkpointer.load(model=self.backbone, optimizer=None, scheduler=None, dataloader=None, path=self.config.load_model)
            # try:
            #     self.backbone, _, _, _, last_epoch, _, _ = self.checkpointer.load(model=self.backbone, optimizer=None, scheduler=None, dataloader=None, path=self.config.load_model)
            #     start_epoch = last_epoch + 1
            # except Exception:
            #     print(f"Tried loading pretrained weights from {self.config.load_model}, but failed.")

        self.backbone = self.backbone.encoder


    def ignore_patch_embed(self, checkpoint, ignore_modules: Optional[list[str]] = None):
        """ Specifies PrithviWxC patch embedding layers and removes them from model checkpoint. """
        ignore_modules = [] if ignore_modules is None else ignore_modules
        ignore_layers = []
        for layer in checkpoint:
            if any(module in layer for module in ignore_modules):
                ignore_layers.append(layer)

        for layer in ignore_layers:
            checkpoint.pop(layer)

        return checkpoint, ignore_layers
    
    def freeze_unused_parameters(self, unused_parameters: list):
        for name, param in self.backbone.named_parameters():        
            if any(unused in name for unused in unused_parameters):
                param.requires_grad = False 

    def freeze_model(self, model: torch.nn.Module, ignore_layers: Optional[list] = None):
        """ Freeze given model for all layers expect the ones specified in ignore_layers. """
        ignore_layers = [] if ignore_layers is None else ignore_layers
        for name, param in model.named_parameters():        
            if name not in (ignore_layers):
                param.requires_grad = False


class ClimateDownscaleFinetuneModel(FinetuneWrapper):

    def __init__(
            self,
            embedding: torch.nn.Module,
            upscale: torch.nn.Module,
            backbone: torch.nn.Module,
            head: torch.nn.Module,
            embed_dim_backbone: int,
            n_lats_px_backbone: int,
            n_lons_px_backbone: int,
            patch_size_px_backbone: tuple[int, int],
            mask_unit_size_px_backbone: tuple[int, int],
            input_scalers_mu: torch.tensor,
            input_scalers_sigma: torch.tensor,
            input_scalers_epsilon: float,
            static_input_scalers_mu: torch.Tensor,
            static_input_scalers_sigma: torch.Tensor,
            static_input_scalers_epsilon: float,
            output_scalers_mu: torch.tensor,
            output_scalers_sigma: torch.tensor,
            n_input_timestamps: int = 1,
            embedding_static: Optional[torch.nn.Module] = None,
            n_bins: int = 512,
            return_logits: bool = False,
            residual: str = None,
            residual_connection: bool = False,
            config=None,
        ):
        """ Climate Downscaling Model based on pre-trained backbone. 
        Args:
            embedding: module used to embed input [C, H, W] -> [E, h, w]
            backbone: module that learns the system dynamics (optionally fully trained) [E, h, w] -> [E, h, w]
            head: module to shape output  [E, h, w] -> [O, H, W]
            n_lats_px_backbone: Total latitudes in data. In pixels.
            n_lons_px_backbone: Total longitudes in data. In pixels
            patch_size_px_backbone: Patch size for tokenization. In pixels lat/lon
            mask_unit_size_px_backbone: Size of each mask unit. In pixels lat/lon
            input_scalers_mu: Tensor of size (in_channels,). Used to rescale input
            input_scalers_sigma:Tensor of size (in_channels,). Used to rescale input
            input_scalers_epsilon: Used to rescale input/ define a lower limit on std
            target_scalers_mu: Tensor of shape (in_channels,). Used to rescale output.
            target_scalers_sigma: Tensor of shape (in_channels,). Used to rescale output.
            n_bins: (optional) Used for cross entropy loss
            return_logits: (optional) Used to determine if we cross entropy loss
            residual: (optional) Indicates the residual mode of the model. for regression
                ['climate',  None]
            residual_connection: (optional) Use a skip/residual connection around the model backbone
        """

        super().__init__(backbone, head, config=config)

        self.n_input_timestamps = n_input_timestamps

        self.residual = residual if residual is not None else ''
        self.residual_connection = residual_connection

        self.embedding = embedding
        self.embedding_static = embedding_static

        self.upscale = upscale

        self.conv_after_backbone = nn.Conv2d(
            embed_dim_backbone, 
            embed_dim_backbone, 
            kernel_size=3, 
            stride=1,
            padding='same',
            padding_mode='replicate'
        )

        # Input shape [batch, time x parameter, lat, lon]
        self.input_scalers_mu = torch.nn.Parameter(
            input_scalers_mu.reshape(1, -1, 1, 1), requires_grad=False
        )
        self.input_scalers_sigma = torch.nn.Parameter(
            input_scalers_sigma.reshape(1, -1, 1, 1), requires_grad=False
        )
        self.input_scalers_epsilon = input_scalers_epsilon

        # Static inputs shape [batch, parameter, lat, lon]
        self.static_input_scalers_mu = nn.Parameter(
            static_input_scalers_mu.reshape(1, -1, 1, 1), requires_grad=False
        )
        self.static_input_scalers_sigma = nn.Parameter(
            static_input_scalers_sigma.reshape(1, -1, 1, 1), requires_grad=False
        )
        self.static_input_scalers_epsilon = static_input_scalers_epsilon

        if output_scalers_mu is not None:
            self.output_scalers_mu = torch.nn.Parameter(
                output_scalers_mu.reshape(1, -1, 1, 1), requires_grad=False
            )
        self.output_scalers_sigma = torch.nn.Parameter(
            output_scalers_sigma.reshape(1, -1, 1, 1), requires_grad=False
        )

        assert n_lats_px_backbone % mask_unit_size_px_backbone[0] == 0
        assert n_lons_px_backbone % mask_unit_size_px_backbone[1] == 0
        assert mask_unit_size_px_backbone[0] % patch_size_px_backbone[0] == 0
        assert mask_unit_size_px_backbone[1] % patch_size_px_backbone[1] == 0

        self.local_shape_mu = (
            int(mask_unit_size_px_backbone[0] // patch_size_px_backbone[0]),
            int(mask_unit_size_px_backbone[1] // patch_size_px_backbone[1]),
        )
        self.global_shape_mu = (
            int(n_lats_px_backbone // mask_unit_size_px_backbone[0]),
            int(n_lons_px_backbone // mask_unit_size_px_backbone[1]),
        )
        self.patch_size_px = patch_size_px_backbone

        self.return_logits = return_logits
        if self.return_logits:
            self.to_logits = nn.Conv2d(
                in_channels=n_bins,
                out_channels=n_bins,
                kernel_size=1,
            )

    def swap_masking(self) -> None:
        return  

    #@profile
    def forward(self, batch: dict[str, torch.tensor]):
        """
        Args:
            batch: Dictionary containing the keys 'x', 'y', and 'static'.
                The associated torch tensors have the following shapes:
                x: Tensor of shape [batch, time x parameter, lat, lon]
                y: Tensor of shape [batch, parameter, lat, lon]
                static: Tensor of shape [batch, channel_static, lat, lon]
                climate: Optional tensor of shape [batch, parameter, lat, lon]
        Returns:
            Tensor of shape [batch, parameter, lat, lon].
        """

        B, _, H, W = batch['x'].shape
        # Scale inputs
        x_sep_time = batch['x'].view(B, self.n_input_timestamps, -1, H, W) # [batch, time x parameter, lat, lon] -> [batch, time, parameter, lat, lon]
        x_scale = (x_sep_time - self.input_scalers_mu.view(1, 1, -1, 1, 1)) / ( 
                self.input_scalers_sigma.view(1, 1, -1, 1, 1) + self.input_scalers_epsilon)
        x = x_scale.view(B, -1, H, W) # [batch, time, parameter, lat, lon] -> [batch, time x parameter, lat, lon]
        
        x_static = (batch['static_x'] - self.static_input_scalers_mu) / (
            self.static_input_scalers_sigma + self.static_input_scalers_epsilon
        )

        if self.residual == 'climate':
            # Scale climatology
            climate = (batch['climate_x'] - self.input_scalers_mu) / (
                self.input_scalers_sigma + self.input_scalers_epsilon
            )

            # concat with static in channels dimension
            x_static = torch.cat([x_static, climate], dim=1)

        # tokenization
        if self.embedding_static is None:
            x = torch.cat([x, x_static], dim=1) # combine the inputs and static in channel dimension
            x_shallow_feats = self.embedding(x)  # [batch, time x parameter, lat, lon] -> [batch, emb, lat*scale[0], lon*scale[0]]
        else:
            x_embedded = self.embedding(x) # [batch, time x parameter, lat, lon] -> [batch, emb, lat*scale[0], lon*scale[0]]
            static_embedded = self.embedding_static(x_static)
            x_shallow_feats = x_embedded + static_embedded

        x_upscale = self.upscale(x_shallow_feats)

        x_tokens = (
            x_upscale.reshape(
                B,
                -1,
                self.global_shape_mu[0],
                self.local_shape_mu[0],
                self.global_shape_mu[1],
                self.local_shape_mu[1],
            )
            .permute(0, 2, 4, 3, 5, 1)
            .flatten(3, 4)
            .flatten(1, 2)
        )  # [batch, embed, lat//patch_size, lon//patch_size] -> [batch, global seq, local seq, embed]

        x_deep_feats = self.backbone(x_tokens)  # [batch, global seq, local seq, embed]

        x_deep_feats = x_deep_feats.reshape(
            B,
            self.global_shape_mu[0],
            self.global_shape_mu[1],
            self.local_shape_mu[0],
            self.local_shape_mu[1],
            -1
        ).permute(0, 5, 1, 3, 2, 4)
        x_deep_feats = x_deep_feats.flatten(4, 5).flatten(2, 3)


        x_deep_feats = self.conv_after_backbone(x_deep_feats)

        if self.residual_connection:
            x = x_deep_feats + x_upscale

        x = self.head(x)  # [batch, out_channels, lat*scale[0]*scale[1], lon*scale[0]*scale[1]]

        if self.return_logits:
            x_out = self.to_logits(x)
        elif self.residual == 'climate':
            x_out = self.output_scalers_sigma * x + batch['climate_y']
        else:
            x_out = self.output_scalers_sigma * x + self.output_scalers_mu # [batch, 1, lat_high_res, lon_high_res]
        
        return x_out



import logging
import multiprocessing
import os
import queue
from pathlib import Path
from typing import Optional, Union

import torch
import torch.distributed as dist
from filelock import FileLock, Timeout
from torch.utils.data import Dataset, IterableDataset

from granitewxc.utils.distributed import get_local_world_size, get_node_rank, get_world_size


def cache_file_name(cache_idx: int, cache_size: int) -> str:
    '''
    Args:
        cache_idx: Position within cache.
        cache_size: Size of cache (in objects).
    '''
    if (cache_idx < 0) or (cache_idx >= cache_size):
        raise ValueError('Ensure that 0 <= `cache_idx` < `cache_size`.')

    return f'cache_{str(cache_idx).zfill(len(str(cache_size)))}.pt'


class CachedDataset(IterableDataset):
    '''
    Iterable dataset to continuously sample samples from local cache as maintained by Cache object.

    See https://pytorch.org/docs/stable/data.html#torch.utils.data.IterableDataset for
    documentation on IterableDataset.
    '''

    def __init__(self, cache_dir: Path, cache_size: int, n_samples_per_epoch: int):
        '''
        Args:
            cache_dir: Location of cache.
            cache_size: Size of the cache.
            n_samples_per_epoch: Indicates how many samples should be returned per epoch.
        '''
        self.cache_dir = cache_dir
        self.cache_size = cache_size
        self.n_samples_per_epoch = n_samples_per_epoch

    def __len__(self):
        return self.n_samples_per_epoch

    def __iter__(self):
        g = torch.Generator()
        index_order = torch.randperm(self.cache_size, generator=g).tolist()
        current_index = 0
        samples_returned = 0

        while True:
            cache_idx = index_order[current_index]
            file_name = self.cache_dir / cache_file_name(cache_idx, self.cache_size)

            lock_file_name = file_name.with_suffix('.lock')
            try:
                with FileLock(lock_file_name, timeout=0):
                    if os.path.isfile(file_name):
                        sample = torch.load(file_name)
                    else:
                        sample = None
            except Timeout:
                sample = None

            current_index += 1
            if current_index >= len(index_order):
                current_index = 0

            if sample is not None:
                yield sample
                samples_returned += 1

            if samples_returned >= self.n_samples_per_epoch:
                break


class Cache:
    '''
    Caches slow-to-retrieve dataset locally.

    Usage example:
    ```
    cache = Cache(
        dataset,
        'path/to/cache',
        cache_size=16,
        n_workers=2,
        clean_up=True
    )
    cache.start()

    data_loader = DataLoader(cache.cached_dataset, batch_size=16, num_workers=2)

    for epoch in range(epochs):
        cache.set_epoch(epoch)
        for idx, sample in enumerate(data_loader):
            # ...
    cache.shutdown()
    ```

    See the pytorch DistributedSampler code
    (https://pytorch.org/docs/stable/_modules/torch/utils/data/distributed.html#DistributedSampler)
    regarding the sampling logic.
    '''

    def __init__(
        self,
        dataset: Dataset,
        cache_dir: Union[Path, str],
        cache_size: int = 64,
        n_workers: int = 1,
        num_replicas: Optional[int] = None,
        node_rank: Optional[int] = None,
        seed: int = 0,
        inactive: bool = False,
        clean_up: bool = True,
    ):
        '''
        For DDP and similar, only the process with LOCAL_RANK 0 should be active and
        all others inactive. Use `inactive` for this.

        Args:
            dataset: Dataset to be cached
            cache_dir: Local dataset to use for caching
            cache_size: Number os samples to store in cache
            n_workers: Number of processes to use to fill and refresh cache
            num_replicas: Number of nodes participating in
                distributed training. By default, this is
                `world_size` // `local_world_size` as retrieved from the
                current distributed group. Note that this assumes identical local world
                size on all nodes.
            node_rank: Rank of the current process group. By default, this is retrieved from the
                current distributed group.
            seed: random seed. Should be identical across all processes in the distributed
                group.
            inactive: Indicates that this particular instance should not load any data.
            clean_up: Delete all cached elements at shutdown.
        '''
        if isinstance(cache_dir, str):
            cache_dir = Path(cache_dir)
        if not cache_dir.is_dir():
            raise ValueError(f'{cache_dir} is not a directory.')

        if num_replicas is None:
            if dist.is_available() and dist.is_initialized():
                world_size = get_world_size()
                local_world_size = get_local_world_size()
                num_replicas = world_size // local_world_size
            else:
                num_replicas = 1
        if node_rank is None:
            if dist.is_available() and dist.is_initialized():
                node_rank = get_node_rank()
            else:
                node_rank = 0
        if node_rank >= num_replicas or node_rank < 0:
            raise ValueError(
                f"Invalid node_rank {node_rank}, node_rank should be in the interval [0, {num_replicas - 1}]"
            )

        self.dataset = dataset
        self.cache_dir = cache_dir
        self.cache_size = cache_size
        self.n_workers = n_workers
        self.num_replicas = num_replicas
        self.node_rank = node_rank
        self.seed = seed
        self.inactive = inactive
        self.clean_up = clean_up

        self.epoch = 0

        if self.inactive:
            self.queues = []
            self.log_queues = []
            self.workers = []
        else:
            self.queues = [multiprocessing.Queue(maxsize=len(dataset)) for i in range(n_workers)]
            self.log_queues = [multiprocessing.Queue(maxsize=1) for i in range(n_workers)]
            self.workers = []

        self.cached_dataset = CachedDataset(
            cache_dir=self.cache_dir, cache_size=cache_size, n_samples_per_epoch=len(dataset)
        )

    def set_epoch(self, epoch: int) -> None:
        '''
        Sets the epoch. This is relevant for sampling purposes.

        Args:
            epoch: Current epoch.
        '''

        self.epoch = epoch

        if not self.inactive:
            self.queue_indices()

    def queue_indices(self) -> None:
        '''
        Queue new indices for caching. The new set of indices will slowly replace previously cached data.
        '''

        if self.inactive:
            return

        if len(self.workers) == 0:
            raise RuntimeError('Start the workers before queueing indices.')

        for q in self.queues:
            q.put(-1)

        g = torch.Generator()
        g.manual_seed(self.seed + self.epoch)
        indices = torch.randperm(len(self.dataset), generator=g).tolist()
        indices = indices[self.node_rank : len(self.dataset) : self.num_replicas]

        for worker_idx in range(self.n_workers):
            for sample_idx in indices[worker_idx :: self.n_workers]:
                self.queues[worker_idx].put(sample_idx)

    def samples_retrieved(self) -> int:
        '''
        Obtain log data from workers.
        Returns:
            For active workers, number of samples retrieved since last call to `set_epoch` or `queue_indices`.
            For inactive workers, 0.
        '''

        if self.inactive:
            return 0

        if len(self.workers) == 0:
            raise RuntimeError('Start the workers before requesting logs indices.')

        # Ask all workers to submit logging information
        for q in self.queues:
            q.put(-2)

        samples_retrieved = [q.get() for q in self.log_queues]
        samples_retrieved = sum(samples_retrieved)

        return samples_retrieved

    def elements_cached(self) -> int:
        '''
        Returns number of elements currently cached.

        Returns:
            Number of cache slots currently occupied.
        '''
        n_cached = 0

        for cache_idx in range(self.cache_size):
            file_name = self.cache_dir / cache_file_name(cache_idx, self.cache_size)

            lock_file_name = file_name.with_suffix('.lock')
            try:
                with FileLock(lock_file_name, timeout=0):
                    if os.path.isfile(file_name):
                        n_cached += 1
            except Timeout:
                pass

        return n_cached

    def half_full(self) -> bool:
        '''
        Check if at least half the cache slots are occupied.

        Returns:
            True if at least half the cache slots are occupied. Otherwise false.
        '''

        return self.elements_cached() >= self.cache_size / 2.0

    def start(self) -> None:
        '''
        Start the worker processes.
        '''

        for worker_idx, (q, log_q) in enumerate(zip(self.queues, self.log_queues)):
            p = multiprocessing.Process(
                target=Cache.worker,
                kwargs={
                    'dataset': self.dataset,
                    'cache_dir': self.cache_dir,
                    'cache_size': self.cache_size,
                    'index_queue': q,
                    'worker_idx': worker_idx,
                    'n_workers': self.n_workers,
                    'clean_up': self.clean_up,
                    'log_queue': log_q,
                },
                daemon=True,
            )
            p.start()
            self.workers.append(p)

    def shutdown(self) -> None:
        if len(self.workers) > 0:
            for q in self.queues:
                q.put(None, block=True, timeout=None)
            for p in self.workers:
                p.join()

    @staticmethod
    def worker(
        dataset: Dataset,
        cache_dir: Path,
        cache_size: int,
        index_queue: multiprocessing.Queue,
        worker_idx: int,
        n_workers: int,
        clean_up: bool = True,
        log_queue: Optional[multiprocessing.Queue] = None,
    ) -> None:
        '''
        Cache worker. Communicates with main process via queue `queue`. Main worker process sends indices of samples to
        retrieve to queue. The worker will store these in the order in which retrieved. Duplicate indices will be dropped.
        The worker continuously loops through the local list of indices as well as cache positions, retrieves the data
        and stores it in cache.

        Args:
            dataset: Pytorch dataset to cache.
            cache_dir: Directory to use for cache.
            cache_size: Maximum number of elements stored in cache.
            index_queue: Multiprocessing queue to communicate with worker.
            worker_idx: Identifies this worker vs. others.
            n_workers: Number of workers on node.
            clean_up: Set to bool to remove all files from cache at termination.
            log_queue: Queue to return logging information to the main process.
        '''
        dataset_indices = []
        current_dataset_idx = 0
        samples_retrieved = 0
        terminate_worker = False
        # Assure worker_idx < cache_size
        if worker_idx >= cache_size:
            raise ValueError('There should not be more workers than cache slots.')
        cache_idx = worker_idx

        # Main worker loop
        while True:
            # Retrieve all elements from queue
            while True:
                try:
                    idx = index_queue.get(block=False)
                except queue.Empty:
                    break

                if idx is None:
                    terminate_worker = True
                    break
                elif idx == -2:
                    if log_queue is not None:
                        try:
                            log_queue.put(samples_retrieved, block=False)
                        except queue.Full:
                            pass
                    samples_retrieved = 0
                elif idx == -1:
                    dataset_indices = []
                    current_dataset_idx = 0
                elif idx in dataset_indices:
                    continue
                else:
                    dataset_indices.append(idx)
                    # Given that we are checking for duplicates above, we should never have too many indices.
                    # So this should never be triggered.
                    if len(dataset_indices) > len(dataset):
                        raise ValueError(
                            'Attempting to keep track of more indices than present in dataset.'
                        )

            # Check if asked to terminate
            if terminate_worker:
                logging.info('Terminating cache worker')
                break

            # Check if there is anything to query
            if len(dataset_indices) == 0:
                continue

            if current_dataset_idx >= len(dataset_indices):
                current_dataset_idx = 0
            if cache_idx >= cache_size:
                cache_idx = worker_idx

            # Retrieve and store data
            sample_idx = dataset_indices[current_dataset_idx]
            sample = dataset[sample_idx]
            file_name = cache_dir / cache_file_name(cache_idx, cache_size)

            lock_file_name = file_name.with_suffix('.lock')
            try:
                with FileLock(lock_file_name, timeout=-1):
                    torch.save(sample, file_name)
                samples_retrieved += 1
            except Timeout:
                raise RuntimeError('Blocking call leads to timeout.')
            current_dataset_idx += 1
            cache_idx += n_workers

        # Clean-up. Simply remove all files.
        if clean_up:
            for cache_idx in range(cache_size):
                file_name = cache_dir / cache_file_name(cache_idx, cache_size)
                try:
                    os.remove(file_name)
                except IOError:
                    pass
                try:
                    os.remove(file_name.with_suffix('.lock'))
                except IOError:
                    pass



import numpy as np
from typing import Optional, Callable

import torch
from torch.utils.data import Dataset

from PrithviWxC.dataloaders.merra2 import Merra2Dataset

class Merra2DownscaleDataset(Dataset):
    """For Climate downscaling Input data has low resolution, and target data has high resolution
    Further information on downscaling: https://link.springer.com/article/10.1007/s00382-022-06343-9#Abs1
    """

    def __init__(
        self,
        data_path_surface: str,
        data_path_vertical: str,
        output_vars: list[str],
        input_surface_vars: Optional[list[str]] = None,
        input_static_surface_vars: Optional[list[str]] = None,
        input_vertical_vars: Optional[list[str]] = None,
        input_levels: Optional[list[float]] = None,
        time_range: slice = None,
        climatology_path_surface: Optional[str] = None,
        climatology_path_vertical: Optional[str] = None,
        transforms: list[Callable] = [],
        n_input_timestamps = 1,
        **kwargs,
    ):
        """
        Args:
            data_path: Path to data of a downscale model (high res data)
            n_input_timestamps: Number of time steps in input
            input_size_level: Number of levels in input
            input_size_lon: Number of lon in input
            input_size_lat: Number of lat in input
            input_surface_vars: input surface variables
            input_upper_level_vars: input upper level variables
        """

        self.input_surface_vars = input_surface_vars
        self.input_static_surface_vars = input_static_surface_vars
        self.input_vertical_vars = input_vertical_vars
        self.input_levels = input_levels if input_levels is not None else list()
        self.output_vars = output_vars # assume surface vars
        self.transforms = transforms

        self.surface_vars = input_surface_vars
        # for cases output parameter in not included in surface
        # maintains input order which is important because of how the scalers are fetched
        # won't need to maintain order if scalers are part of Dataset class 
        for out_var in self.output_vars:
            if out_var not in self.surface_vars:
                self.surface_vars.append(out_var)

        input_times = [int(timestamp) for timestamp in np.linspace(n_input_timestamps*-3, -3, n_input_timestamps)]

        self.dataset = Merra2Dataset(
            data_path_surface = data_path_surface,
            data_path_vertical = data_path_vertical,
            climatology_path_surface = climatology_path_surface,
            climatology_path_vertical = climatology_path_vertical,
            surface_vars = self.surface_vars,
            static_surface_vars = self.input_static_surface_vars,
            vertical_vars = self.input_vertical_vars,
            levels = self.input_levels,
            lead_times = [0],
            time_range = time_range,
            input_times = input_times,
        )

        self.n_input_timestamps = n_input_timestamps

        self.num_datapoints = self.dataset.__len__()

    @staticmethod
    def _subset_idx_map(x, x_sub) -> torch.Tensor:
        """
        e.g.
        x = ['t', 'u', 'v']
        x_sub = ['v', 'u']
        returns [2, 1] which is the index of target_vars in input_vars
        """

        return torch.tensor([x.index(var) for var in x_sub])


    def get_data(self, index) -> dict[torch.Tensor | int]:
        """
        Returns a dictionary with keys `sur_static`, `ulv_static`, `sur_vals`, `ulv_vals`, `sur_tars`, `ulv_tars`.
        Each individual component is a numpy array. They generally have the shape `feature, dim_0, dim_1, dim_2, ...`.
        Which dimensions are present depends on the particular tensor.
        The dimension ordering is time -> level -> lat -> lon.

        Args:
            index: Index of the sample returned.
        Returns:
            Dictionary with keys 'sur_static', 'sur_vals', 'sur_tars', 'ulv_static', 'ulv_vals', 'ulv_tars',
                'lead_time'. For each, the value is as follows:
            sur_static: Numpy array of shape (7, lat, lon). For each pixel (lat, lon),
                the first dimension indexes sin(lat), cos(lon), sin(lon), cos(doy), sin(doy), cos(hod), sin(hod).
                Where doy is the day of the year [1, 366] and hod the hour of the day [0, 23].
            sur_vals: Torch tensor of shape (time, parameter, lat, lon).
            sur_tars: Torch tensor of shape (parameter, lat, lon).
            ulv_static: Numpy array of shape (4, level, lat, lon). For each pixel (level, lat, lon),
                the first dimension indexes level, sin(lat), cos(lon), sin(lon).
            ulv_vals: Torch tensor of shape (time, parameter, level, lat, lon).
            ulv_tars: Torch tensor of shape (time, parameter, level, lat, lon).
            lead_time: Integer. Optional
        """
        data = self.dataset.__getitem__(index)

        x_sur = data['sur_vals'].index_select(index=self._subset_idx_map(self.surface_vars, self.input_surface_vars), dim=0) # parameters x time x lat x lon
        x_upl = data['ulv_vals'] # parameters x level x time x lat x lon

        static = data['sur_static']

        x = torch.cat(
            [
                x_sur.reshape(-1, x_sur.shape[-3], x_sur.shape[-2], x_sur.shape[-1]), #  parameters,  time, lat,  lon
                x_upl.reshape(-1, x_sur.shape[-3], x_upl.shape[-2], x_upl.shape[-1]), # parameters x level, time, lat,  lon
            ],
            dim=0,
        ) # parameters x level, time, lat,  lon

        x = x.permute(1, 0, 2, 3).reshape(-1, x_sur.shape[-2], x_sur.shape[-1]) # time x parameters x level,  lat,  lon

        y = data['sur_vals'].index_select( # parameters x time x lat x lon
            index=self._subset_idx_map(self.surface_vars, self.output_vars), dim=0
        )
        
        if self.n_input_timestamps > 1:
            # parameters, time, lat, lon -> parameters, 1, lat, lon
            y = y[:, -1:]

        y = y.reshape(
            -1, data['sur_vals'].shape[-2], data['sur_vals'].shape[-1]
        ) # parameters, lat, lon

        ret = {
            'x': x,
            'y': y,
            'static': static,
        }

        if 'sur_climate' in data:
            climate_sur = data['sur_climate'].index_select(index=self._subset_idx_map(self.surface_vars, self.input_surface_vars), dim=0) # parameters, lat, lon
            climate_upl = data['ulv_climate'] # parameters, level, lat, lon

            ret['climate_x'] = torch.cat(
                [
                    climate_sur,
                    climate_upl.reshape(-1, climate_upl.shape[-2], climate_upl.shape[-1]),
                ],
                dim=0,
            )

            ret['climate_y'] = data['sur_climate'].index_select(
                index=self._subset_idx_map(self.surface_vars, self.output_vars), dim=0
                ).reshape(
                    -1, data['sur_climate'].shape[-2], data['sur_climate'].shape[-1]
            )

        return ret
    
    
    def __getitem__(self, index) -> dict[torch.Tensor | int]:
        """
        Returns a dictionary with keys: [x, x_static, y, y_static, climate].
        Each individual component is a numpy array. They generally have the shape `feature, dim_0, dim_1, dim_2, ...`.
        Which dimensions are present depends on the particular tensor.
        The dimension ordering is time -> level -> lat -> lon.

        Args:
            index: Index of the sample returned.
        Returns:
            Dictionary with keys [x, x_static, y, y_static, climate],
        """

        sample = self.get_data(index)
            
        for t in self.transforms:
            sample = t(sample)
        
        return sample


    def __str__(self) -> str:
        return (
            f"DownscaleDataset("
            f"Recs: {len(self)}, "
            f"Dataset: {str(self.dataset)}, "
        )


    def __len__(self) -> int:
        return self.dataset.__len__()
        



from typing import Optional
from logging import Logger

import numpy as np
import torch

from PrithviWxC.model import PrithviWxCEncoderDecoder, SWINShift

from granitewxc.utils.config import ExperimentConfig
from granitewxc.utils import distributed
from granitewxc.utils.data import (assemble_input_scalers,
                                       assemble_target_scalers,
                                       assemble_output_scalers,
                                       assemble_static_input_scalers,
                                       assemble_climate_scalers)
from granitewxc.models.finetune_model import PatchEmbed, ClimateDownscaleFinetuneModel
from granitewxc.decoders.downscaling import ConvEncoderDecoder


def get_scalers(config: ExperimentConfig):
    """
    calls assemble scalers func. depending on the dataset
    """
    # Input and target scalers
    if config.data.type == 'merra2':

        input_mu, input_sigma = assemble_input_scalers(config)

        if config.model.__dict__.get('residual', '') == 'climate':
            target_variance = assemble_climate_scalers(config)
            target_sigma = torch.sqrt(target_variance)
            target_mu = None
        else:
            target_mu, target_sigma = assemble_target_scalers(config)

        input_static_mu, input_static_sigma = assemble_static_input_scalers(config)

    else:
        raise ValueError(f'{config.data.type} is not a valid config.data.type')

    return dict(
        input_mu=input_mu,
        input_sigma=input_sigma,
        input_static_mu=input_static_mu,
        input_static_sigma=input_static_sigma,
        target_mu=target_mu,
        target_sigma=target_sigma,
    )


def get_merra2_embedding_module(config: ExperimentConfig,):
    """
    Initializes the object required to embed merra2 data
    """

    n_parameters = (len(config.data.input_surface_vars) + len(config.data.input_levels) * len(
        config.data.input_vertical_vars))


    patch_embedding = PatchEmbed(
        patch_size=config.model.downscaling_patch_size,
        channels=n_parameters * config.data.n_input_timestamps,
        embed_dim=config.model.downscaling_embed_dim,
    )

    n_static_parameters = config.model.num_static_channels + len(config.data.input_static_surface_vars)
    if config.model.residual == 'climate':
        n_static_parameters += n_parameters

    patch_embedding_static = PatchEmbed(
        patch_size=config.model.downscaling_patch_size,
        channels=n_static_parameters,
        embed_dim=config.model.downscaling_embed_dim,
    )

    return patch_embedding, patch_embedding_static


def get_merra2_upscaling_module(config: ExperimentConfig,):
    """
    Initializes the object required to embed merra2 data
    """

    n_parameters = (len(config.data.input_surface_vars) + len(config.data.input_levels) * len(
        config.data.input_vertical_vars))


    patch_embedding = ConvEncoderDecoder(
        in_channels=n_parameters * config.data.n_input_timestamps,
        channels=config.model.encoder_decoder_conv_channels,
        out_channels=config.model.embed_dim,
        kernel_size=config.model.encoder_decoder_kernel_size_per_stage[0],
        scale=config.model.encoder_decoder_scale_per_stage[0],
        upsampling_mode=config.model.encoder_decoder_upsampling_mode,
    )

    n_static_parameters = config.model.num_static_channels + len(config.data.input_static_surface_vars)
    if config.model.residual == 'climate':
        n_static_parameters += n_parameters

    patch_embedding_static = ConvEncoderDecoder(
        in_channels=n_static_parameters,
        channels=config.model.encoder_decoder_conv_channels,
        out_channels=config.model.embed_dim,
        kernel_size=config.model.encoder_decoder_kernel_size_per_stage[0],
        scale=config.model.encoder_decoder_scale_per_stage[0],
        upsampling_mode=config.model.encoder_decoder_upsampling_mode,
    )

    return patch_embedding, patch_embedding_static


def get_backbone(config) -> PrithviWxCEncoderDecoder:

    print("Encoder shifting: {}".format(config.model.encoder_shift))

    if config.model.encoder_shift:
        n_lats_px=int(config.data.input_size_lat * np.prod(config.model.encoder_decoder_scale_per_stage[0]))
        n_lons_px=int(config.data.input_size_lon * np.prod(config.model.encoder_decoder_scale_per_stage[0]))
        local_shape_mu = (
            config.mask_unit_size[0] // config.model.token_size[0],
            config.mask_unit_size[1] // config.model.token_size[1],
        )
        global_shape_mu = (
            n_lats_px // config.mask_unit_size[0],
            n_lons_px // config.mask_unit_size[1],
        )
        e_shifter = SWINShift(
            config.mask_unit_size,
            global_shape_mu,
            local_shape_mu,
            config.model.token_size,
            n_context_tokens=0,
        )
    else:
        e_shifter = None

    return PrithviWxCEncoderDecoder(
        embed_dim=config.model.embed_dim,
        n_blocks=config.model.n_blocks_encoder,
        mlp_multiplier=config.model.mlp_multiplier,
        n_heads=config.model.n_heads,
        dropout=config.model.dropout_rate,
        drop_path=config.model.drop_path,
        shifter=e_shifter,
    )


def get_finetune_model(config: ExperimentConfig, logger: Optional[Logger] = None) -> torch.nn.Module:
    """
    Args:
        config: Experiment configuration. Contains configuration parameters for model.
        logger: wandb Run object as returned by wandb.init.
    Returns:
        The configured model.
    """

    if distributed.is_main_process():
        print("Creating the model.")

    #########################################################
    # 0. Setup parameters/scalers
    #########################################################
    # set default kernel size
    if 'encoder_decoder_kernel_size_per_stage' not in config.model.__dict__:
        config.model.encoder_decoder_kernel_size_per_stage = [[3]*len(inner) for inner in config.model.encoder_decoder_scale_per_stage]

    n_output_parameters = len(config.data.output_vars)
    if config.model.__dict__.get('loss_type', 'patch_rmse_loss')=='cross_entropy':
        if config.model.__dict__.get('cross_entropy_bin_width_type', 'uniform') == 'uniform':
            n_output_parameters = config.model.__dict__.get('cross_entropy_n_bins', 512)
        else:
            n_output_parameters = len(np.load(config.model.cross_entropy_bin_boundaries_file)) + 1

    scalers = get_scalers(config)

    #########################################################
    # 1. Patch Embedding/Shallow Feature Extraction
    #########################################################
    if config.data.type == 'merra2':  # merra2
        embedding, embedding_static = get_merra2_embedding_module(config)
    else:
        raise ValueError(f'{config.data.type} is not a valid config.data.type')

    #########################################################
    # 2. Upscale before FM
    # Keep token resolution similar to trained backbone
    #########################################################
    upscale = ConvEncoderDecoder(
        in_channels=config.model.downscaling_embed_dim,
        channels=config.model.encoder_decoder_conv_channels,
        out_channels=config.model.embed_dim,
        kernel_size=config.model.encoder_decoder_kernel_size_per_stage[0],
        scale=config.model.encoder_decoder_scale_per_stage[0],
        upsampling_mode=config.model.encoder_decoder_upsampling_mode,
    )


    #########################################################
    # 3. FM/Deep Feature Extraction
    #########################################################

    backbone = get_backbone(config)

    #########################################################
    # 4. Upscale after FM
    #########################################################
    if config.model.encoder_decoder_type == 'conv':
        head = ConvEncoderDecoder(
                in_channels=config.model.embed_dim,
                channels=config.model.encoder_decoder_conv_channels,
                out_channels=n_output_parameters,
                kernel_size=config.model.encoder_decoder_kernel_size_per_stage[1],
                scale=config.model.encoder_decoder_scale_per_stage[1],
                upsampling_mode=config.model.encoder_decoder_upsampling_mode,
        )
    else:
        raise NotImplementedError(f"Head type {config.model.encoder_decoder_type} not implemented.")

    #########################################################
    # 5. Putting it all together
    #########################################################
    model = ClimateDownscaleFinetuneModel(
        embedding=embedding,
        embedding_static=embedding_static,
        upscale=upscale,
        backbone=backbone,
        head=head,
        input_scalers_mu=scalers['input_mu'],
        input_scalers_sigma=scalers['input_sigma'],
        input_scalers_epsilon=1e-6,
        static_input_scalers_mu=scalers['input_static_mu'],
        static_input_scalers_sigma=scalers['input_static_sigma'],
        static_input_scalers_epsilon=1e-6,
        output_scalers_mu=scalers['target_mu'],
        output_scalers_sigma=scalers['target_sigma'],
        n_input_timestamps=config.data.n_input_timestamps,
        embed_dim_backbone=config.model.embed_dim,
        n_lats_px_backbone=int(config.data.input_size_lat * np.prod(config.model.encoder_decoder_scale_per_stage[0])),
        n_lons_px_backbone=int(config.data.input_size_lon * np.prod(config.model.encoder_decoder_scale_per_stage[0])),
        patch_size_px_backbone=config.model.token_size,
        mask_unit_size_px_backbone=config.mask_unit_size,
        n_bins=n_output_parameters,
        return_logits=config.model.__dict__.get('loss_type')=='cross_entropy',
        residual=config.model.__dict__.get('residual', None),
        residual_connection=config.model.__dict__.get('residual_connection', False),
        config=config,
    )

    if distributed.is_main_process():
        total_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
        if logger is not None:
            logger.info(f"--> model has {total_params:,.0f} params.")
    return model



import os
from functools import partial
from typing import Optional
import logging
from packaging.version import parse as parse_version

import numpy as np
import torch
import xarray as xr
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler

if parse_version(xr.__version__) >= parse_version('2023.11.0'):
    XARRAY_TO_DATAARRAY = True
else:
    XARRAY_TO_DATAARRAY = False

from granitewxc.utils.cache import Cache
from granitewxc.utils.distributed import get_local_rank
from granitewxc.utils import distributed
from granitewxc.utils.config import ExperimentConfig
from granitewxc.datasets.merra2 import Merra2DownscaleDataset

def torch_interpolate_3d(x, size=None, scale_factor=None, mode='bilinear', align_corners=None):
    x = torch.unsqueeze(x, 0)
    x = F.interpolate(x, size=size, scale_factor=scale_factor, mode=mode, align_corners=align_corners)
    x = torch.squeeze(x, 0)
    return x


def crop(sample: dict[str, torch.Tensor], crop_lat: list[int], crop_lon: list[int]):

    for k, v in sample.items():
        sample[k] = v[..., crop_lat[0]:crop_lat[1], crop_lon[0]:crop_lon[1]]
    
    return sample


def transform(sample, size=(35, 35)):

    ret = dict()

    retain_keys = dict(
        y='y',
        static='static_y',
        climate_y='climate_y',
    )

    for k_orig, k_new in retain_keys.items():
        if k_orig in sample:
            ret[k_new] = sample[k_orig]

    interpolate_keys = dict(
        x='x',
        static='static_x',
        climate_x='climate_x'
    )

    for k_orig, k_new in interpolate_keys.items():
        if k_orig in sample:
            ret[k_new] = torch_interpolate_3d(
                sample[k_orig], size=size, mode='bilinear', align_corners=True
            )

    return ret


def smoothen(sample):
    padding_mode = 'replicate'

    for k in ['x', 'static_x', 'climate_x']:
        if k in sample:
            n_dims = sample[k].shape[0]

            kernel = torch.ones((3, 3))
            kernel = kernel / kernel.sum()
            kernel = torch.einsum(
                'ij,kl->ijkl', 
                torch.eye(n_dims), kernel
            )

            sample[k] = F.conv2d(
                F.pad(sample[k], (1, 1, 1, 1), mode=padding_mode),
                kernel, padding='valid'
            )

    return sample


def preproc_downscale_merra2(
        batch: list[dict],
        input_padding: dict[tuple[int]],
        target_padding: dict[tuple[int]],
    ) -> dict:
    """
    inputs and targets can have different parameters 
    Args:
        batch: List of training samples. Each sample should be a dictionary with keys 'sur_static', 'sur_vals',
            'sur_tars', 'ulv_static', 'ulv_vals', 'ulv_tars', 'lead_time'.
            The tensors have shape:
                sur_static: Numpy array of shape (3, lat, lon). For each pixel (lat, lon),
                            the first dimension indexes sin(lat), cos(lon), sin(lon).
                sur_vals: Torch tensor of shape (parameter, time, lat, lon).
                sur_tars: Torch tensor of shape (parameter, time, lat, lon).
                ulv_static: Numpy array of shape (4, level, lat, lon). For each pixel (level, lat, lon),
                            the first dimension indexes level, sin(lat), cos(lon), sin(lon).
                ulv_vals: Torch tensor of shape (parameter, level, time, lat, lon).
                ulv_tars: Torch tensor of shape (parameter, level, time, lat, lon).
                sur_climate: Torch tensor of shape (parameter, lat, lon)
                ulv_climate: Torch tensor of shape (parameter, level, lat, lon)
                lead_time: Integer.
        input_padding: Padding values for input data. Dictionary with keys 'level', 'lat', 'lon. For each,
            the value is a tuple of length 2 indicating padding at the start and end of the relevant dimension.
        target_padding: Padding values for target data (same structure as *input_padding*)
    Returns:
        Dictionary with keys 'x', 'y', 'lead_time' and 'static'. All are torch tensors. Shapes are as follows:
            x: batch, time, parameter, lat, lon
            y: batch, parameter, lat, lon
            static: batch, parameter, lat, lon
            lead_time: batch
        Here, for x and y, 'parameter' is [surface parameter, upper level parameter x level]
    """
    if not set(batch[0].keys()).issuperset(
        ['sur_static', 'sur_vals', 'ulv_static', 'ulv_vals', 'lead_time']
    ):
        raise ValueError('Missing essential keys.')
    if not set(
        [
            'sur_static',
            'sur_vals',
            'sur_tars',
            'ulv_static',
            'ulv_vals',
            'ulv_tars',
            'sur_climate',
            'ulv_climate',
            'lead_time',
        ]
    ).issuperset(batch[0].keys()):
        raise ValueError('Unexpected keys in batch.')
    if set(batch[0].keys()).intersection(set(['ulv_tars', 'sur_tars'])) == set():
        raise ValueError(f"Need or both of {set('ulv_tars', 'sur_tars')}")

    # Bring all tensors from the batch into a single tensor
    upl_x = torch.empty((len(batch), *batch[0]['ulv_vals'].shape))
    sur_x = torch.empty((len(batch), *batch[0]['sur_vals'].shape))

    upl_y = torch.empty((len(batch), *batch[0]['ulv_tars'].shape)) if 'uvl_tars' in batch[0].keys() else None
    sur_y = torch.empty((len(batch), *batch[0]['sur_tars'].shape)) if 'sur_tars' in batch[0].keys() else None

    sur_sta = torch.empty((len(batch), *batch[0]['sur_static'].shape))
    upl_sta = torch.empty((len(batch), *batch[0]['ulv_static'].shape))

    lead_time = torch.empty((len(batch),), dtype=torch.float32)

    for i, rec in enumerate(batch):
        sur_x[i] = torch.Tensor(rec['sur_vals'])
        upl_x[i] = torch.Tensor(rec['ulv_vals'])

        if sur_y is not None: sur_y[i] = torch.Tensor(rec['sur_tars'])
        if upl_y is not None: upl_y[i] = torch.Tensor(rec['ulv_tars'])

        sur_sta[i] = torch.Tensor(rec['sur_static'])
        upl_sta[i] = torch.Tensor(rec['ulv_static'])

        lead_time[i] = rec['lead_time']

    upl_x = upl_x.permute((0, 3, 1, 2, 4, 5))  # (batch, parameter, level, time, lat, lon) -> (batch, time, parameter, level, lat, lon)
    upl_x = upl_x.flatten(1, 2)  # [batch, time, parameter, lat_low_res, lon_low_res] -> [batch, time x parameter, lat_low_res, lon_low_res]
    if upl_y is not None:
        upl_y = upl_y.permute((0, 3, 1, 2, 4, 5))
        upl_y = upl_y.flatten(1, 2)

    sur_x = sur_x.permute((0, 2, 1, 3, 4))  # (batch, parameter, time, lat, lon) -> (batch, time, parameter, lat, lon)
    sur_x = sur_x.flatten(1, 2)  # [batch, time, parameter, lat_low_res, lon_low_res] -> [batch, time x parameter, lat_low_res, lon_low_res]
    if sur_y is not None:
        sur_y = sur_y.permute((0, 2, 1, 3, 4))
        sur_y = sur_y.flatten(1, 2)

    # Pad
    input_padding_2d = (*input_padding['lon'], *input_padding['lat'])
    input_padding_3d = (*input_padding['lon'], *input_padding['lat'], *input_padding['level'])

    target_padding_2d = (*target_padding['lon'], *target_padding['lat'])
    target_padding_3d = (*target_padding['lon'], *target_padding['lat'], *target_padding['level'])

    sur_x = torch.nn.functional.pad(sur_x, input_padding_2d, mode='constant', value=0)
    upl_x = torch.nn.functional.pad(upl_x, input_padding_3d, mode='constant', value=0)
    if sur_y is not None: sur_y = torch.nn.functional.pad(sur_y, target_padding_2d, mode='constant', value=0)
    if upl_y is not None: upl_y = torch.nn.functional.pad(upl_y, target_padding_3d, mode='constant', value=0)
    sur_sta = torch.nn.functional.pad(sur_sta, target_padding_2d, mode='constant', value=0)
    upl_sta = torch.nn.functional.pad(upl_sta, target_padding_3d, mode='constant', value=0)

    # We stack along the combined parameter x level dimension
    x = torch.cat(
        [
            sur_x,
            upl_x.reshape(*upl_x.shape[:1], upl_x.shape[1] * upl_x.shape[2], *upl_x.shape[3:]),
        ],
        dim=1,
    )

    if upl_y is not None and sur_y is not None:
        y = torch.cat(
            [
                sur_y,
                upl_y.reshape(upl_y.shape[0], upl_y.shape[1] * upl_y.shape[2], *upl_y.shape[3:]),
            ],
            dim=1,
        )
    elif upl_y is None:
        y = sur_y
    elif sur_y is None:
        y = upl_y.reshape(upl_y.shape[0], upl_y.shape[1] * upl_y.shape[2], *upl_y.shape[3:])
    else:  # should not reach this part of the code
        raise ValueError("No targets in data")

    return_value = {
        'x': x,
        'y': y,
        'lead_time': lead_time,
        'static': sur_sta,
    }

    return return_value


def get_dataloaders_merra2(config: ExperimentConfig) -> tuple[DataLoader, DataLoader, Optional[Cache], Optional[Cache]]:
    """
    Args:
        config: Experiment configuration. Contains configuration parameters for model.
    Returns:
        Tuple of data loaders: (training loader, validation loader).
    """

    ds_kwargs = dict(
        data_path_surface = config.data.data_path_surface,
        data_path_vertical = config.data.data_path_vertical,
        climatology_path_surface = config.data.climatology_path_surface,
        climatology_path_vertical = config.data.climatology_path_vertical,
        input_surface_vars = config.data.input_surface_vars,
        input_static_surface_vars = config.data.input_static_surface_vars,
        input_vertical_vars = config.data.input_vertical_vars,
        input_levels = config.data.input_levels,
        n_input_timestamps = config.data.n_input_timestamps,
        output_vars=config.data.output_vars,
        transforms=_get_transforms(config),
    )

    train_dataset = Merra2DownscaleDataset(
        time_range=(config.data.train_time_range_start, config.data.train_time_range_end),
        **ds_kwargs
    )

    valid_dataset = Merra2DownscaleDataset(
        time_range=(config.data.val_time_range_start, config.data.val_time_range_end),
        **ds_kwargs
    )

    if config.data.cache:
        # We might have to adjust the cache dir ...
        pbs_job_id = os.environ.get('PBS_JOBID')
        if pbs_job_id is not None:
            config.data.cache_dir = os.path.join(config.data.cache_dir, pbs_job_id)

        cache_dir_train = os.path.join(config.data.cache_dir, 'train')
        cache_dir_valid = os.path.join(config.data.cache_dir, 'valid')
        os.makedirs(cache_dir_train, exist_ok=True)
        os.makedirs(cache_dir_valid, exist_ok=True)

        train_cache = Cache(
            train_dataset,
            cache_dir_train,
            cache_size=config.data.cache_size_train,
            n_workers=config.data.cache_num_workers,
            inactive=get_local_rank() != 0,
            clean_up=True,
        )
        valid_cache = Cache(
            valid_dataset,
            cache_dir_valid,
            cache_size=config.data.cache_size_valid,
            n_workers=config.data.cache_num_workers,
            inactive=get_local_rank() != 0,
            clean_up=True,
        )
        train_cache.start()
        valid_cache.start()

        if distributed.is_main_process():
            print(f"Initialized {config.data.cache_dir} as cache.")
    else:
        train_cache = None
        valid_cache = None

    dl_kwargs = dict(
        batch_size=config.batch_size,
        num_workers=config.dl_num_workers,
        prefetch_factor=config.dl_prefetch_size,
        pin_memory=True,
        drop_last=True,
        persistent_workers=True,
    )

    if config.data.cache:
        train_loader = DataLoader(dataset=train_cache.cached_dataset, **dl_kwargs)
        valid_loader = DataLoader(dataset=valid_cache.cached_dataset, **dl_kwargs)
    else:
        train_loader = DataLoader(
            dataset=train_dataset,
            sampler=DistributedSampler(train_dataset, shuffle=True, drop_last=True),
            **dl_kwargs,
        )
        valid_loader = DataLoader(
            dataset=valid_dataset,
            sampler=DistributedSampler(valid_dataset, shuffle=False, drop_last=False),
            **dl_kwargs,
        )

    if distributed.is_main_process():
        print(f"--> Training batches: {len(train_loader):,.0f}")
        print(f"--> Validation batches: {len(valid_loader):,.0f}")
        print(f"--> Training samples: {len(train_dataset):,.0f}")
        print(f"--> Validation samples: {len(valid_dataset):,.0f}")

    return (
        train_loader,
        valid_loader,
        train_cache,
        valid_cache,
    )


def _get_transforms(config: ExperimentConfig):

    if 'crop_lat' in config.data.__dict__ and config.data.crop_lat is not None:
        crop_lat = [config.data.crop_lat[0], -config.data.crop_lat[1]]
    else:
        crop_lat = [0, None]

    if 'crop_lon' in config.data.__dict__ and config.data.crop_lon is not None:
        crop_lon = [config.data.crop_lon[0], -config.data.crop_lon[1]]
    else:
        crop_lon = [0, None]

    transforms = [
        partial(crop, crop_lat=crop_lat, crop_lon=crop_lon),
        partial(transform, size=(config.data.input_size_lat, config.data.input_size_lon)),
    ]
                  

    if config.data.__dict__.get('apply_smoothen', False):
        transforms.append(smoothen)

    return transforms


def get_dataloaders(config: ExperimentConfig, logger: logging.Logger=None) -> tuple[DataLoader, DataLoader]:
    """
    Calls one of:
      get_dataloaders_merra2
      get_dataloaders_cordex
    """
    if config.data.type == 'merra2':
        return get_dataloaders_merra2(config)
    else:
        raise ValueError(f'{config.data.type} is not a valid config.data.type')


def assemble_input_scalers(config: ExperimentConfig) -> tuple[torch.Tensor, torch.Tensor]:
    """
    For merra2 based dataset where surface and upper level variables are specified separately
    """
    _kwargs_open_dataset = {'chunks' : None, 'cache' : True, 'engine' : 'h5netcdf'}

    with (
        xr.open_dataset(
            config.model.input_scalers_surface_path,
            **_kwargs_open_dataset
        ) as musigma_surface,
        xr.open_dataset(
            config.model.input_scalers_vertical_path,
            **_kwargs_open_dataset
        ) as musigma_vertical
    ):
        musigma_surface.load()
        musigma_vertical.load()
        
        mu_surface = musigma_surface[config.data.input_surface_vars].sel(statistic='mu')
        sigma_surface = musigma_surface[config.data.input_surface_vars].sel(statistic='sigma')
        mu_vertical = musigma_vertical[config.data.input_vertical_vars].sel(statistic='mu', lev=config.data.input_levels)
        sigma_vertical = musigma_vertical[config.data.input_vertical_vars].sel(statistic='sigma', lev=config.data.input_levels)

        if XARRAY_TO_DATAARRAY:
            mu_surface = mu_surface.to_dataarray(dim='parameter')
            sigma_surface = sigma_surface.to_dataarray(dim='parameter')
            mu_vertical = mu_vertical.to_dataarray(dim='parameter')
            sigma_vertical = sigma_vertical.to_dataarray(dim='parameter')
        else:
            mu_surface = mu_surface.to_array(dim='parameter')
            sigma_surface = sigma_surface.to_array(dim='parameter')
            mu_vertical = mu_vertical.to_array(dim='parameter')
            sigma_vertical = sigma_vertical.to_array(dim='parameter')
        
        mu = torch.cat(
            (
                torch.from_numpy(mu_surface.values),
                torch.from_numpy(mu_vertical.transpose('parameter', 'lev').values.reshape(-1)),
            ), dim=0
        ).to(dtype=torch.float32)
        
        sigma = torch.cat(
            (
                torch.from_numpy(sigma_surface.values),
                torch.from_numpy(sigma_vertical.transpose('parameter', 'lev').values.reshape(-1)),
            ), dim=0
        ).to(dtype=torch.float32)

        sigma = torch.clamp(sigma, 1e-4, 1e4)

    return mu, sigma


def assemble_target_scalers(config: ExperimentConfig) -> tuple[torch.Tensor, torch.Tensor]:
    """
    For merra2 based dataset where surface and upper level variables are specified separately
    """
    _kwargs_open_dataset = {'chunks' : None, 'cache' : True, 'engine' : 'h5netcdf'}

    with (
        xr.open_dataset(
            config.model.input_scalers_surface_path,
            **_kwargs_open_dataset
        ) as musigma_surface,
    ):
        musigma_surface.load()
        
        mu_surface = musigma_surface[config.data.output_vars].sel(statistic='mu')
        sigma_surface = musigma_surface[config.data.output_vars].sel(statistic='sigma')

        if XARRAY_TO_DATAARRAY:
            mu_surface = mu_surface.to_dataarray(dim='parameter')
            sigma_surface = sigma_surface.to_dataarray(dim='parameter')
        else:
            mu_surface = mu_surface.to_array(dim='parameter')
            sigma_surface = sigma_surface.to_array(dim='parameter')
        
        mu = torch.from_numpy(mu_surface.values).to(dtype=torch.float32)
        sigma = torch.from_numpy(sigma_surface.values).to(dtype=torch.float32)
        sigma = torch.sqrt(torch.clamp(sigma**2, 1e-8, 1e6))

        return mu, sigma


def assemble_static_input_scalers(config: ExperimentConfig) -> tuple[torch.Tensor, torch.Tensor]:
    _kwargs_open_dataset = {'chunks' : None, 'cache' : True, 'engine' : 'h5netcdf'}

    with xr.open_dataset(
            config.model.input_scalers_surface_path,
            **_kwargs_open_dataset
        ) as musigma_surface:
        musigma_surface.load()
        
        mu_surface = musigma_surface[config.data.input_static_surface_vars].sel(statistic='mu')
        sigma_surface = musigma_surface[config.data.input_static_surface_vars].sel(statistic='sigma')

        if XARRAY_TO_DATAARRAY:
            mu_surface = mu_surface.to_dataarray(dim='parameter')
            sigma_surface = sigma_surface.to_dataarray(dim='parameter')
        else:
            mu_surface = mu_surface.to_array(dim='parameter')
            sigma_surface = sigma_surface.to_array(dim='parameter')
        
        mu = torch.cat(
            (
                torch.zeros((config.model.num_static_channels,), dtype=torch.float32),
                torch.from_numpy(mu_surface.values),
            ), dim=0
        ).to(dtype=torch.float32)
        
        sigma = torch.cat(
            (
                torch.ones((config.model.num_static_channels,), dtype=torch.float32),
                torch.from_numpy(sigma_surface.values),
            ), dim=0
        ).to(dtype=torch.float32)

        sigma = torch.clamp(sigma, 1e-4, 1e4)

        return mu, sigma


def assemble_output_scalers(config: ExperimentConfig) -> torch.Tensor:
    _kwargs_open_dataset = {'chunks' : None, 'cache' : True, 'engine' : 'h5netcdf'}

    if config.model.residual == 'none':
        _, sigma = assemble_input_scalers(config)
        variances = sigma**2
        return variances

    with (
        xr.open_dataset(
            config.model.output_scalers_surface_path,
            **_kwargs_open_dataset
        ) as scaler_surface,
        xr.open_dataset(
            config.model.output_scalers_vertical_path,
            **_kwargs_open_dataset
        ) as scaler_vertical
    ):
        scaler_surface = scaler_surface[config.data.input_surface_vars]
        scaler_vertical = scaler_vertical[config.data.input_vertical_vars].sel(lev=config.data.input_levels)
        
        if XARRAY_TO_DATAARRAY:
            scaler_surface = scaler_surface.to_dataarray(dim='parameter')
            scaler_vertical = scaler_vertical.to_dataarray(dim='parameter')
        else:
            scaler_surface = scaler_surface.to_array(dim='parameter')
            scaler_vertical = scaler_vertical.to_array(dim='parameter')

        variances = torch.cat(
            (
                torch.from_numpy(scaler_surface.values),
                torch.from_numpy(scaler_vertical.transpose('parameter', 'lev').values.reshape(-1)),
            ),
            dim=0
        ).to(dtype=torch.float32)

    # Looking through the numbers, we have values as extreme as 1e-59 and 7e6.
    variances = torch.clamp(variances, 1e-7, 1e7)

    return variances


def assemble_climate_scalers(config) -> torch.Tensor:
    _kwargs_open_dataset = {'chunks' : None, 'cache' : True, 'engine' : 'h5netcdf'}

    with (
        xr.open_dataset(
            config.model.output_scalers_surface_path,
            **_kwargs_open_dataset
        ) as scaler_surface
    ):
        scaler_surface = scaler_surface[config.data.output_vars]

        if XARRAY_TO_DATAARRAY:
            scaler_surface = scaler_surface.to_dataarray(dim='parameter')
        else:
            scaler_surface = scaler_surface.to_array(dim='parameter')

        variances = torch.from_numpy(scaler_surface.values).to(dtype=torch.float32)

    # Looking through the numbers, we have values as extreme as 1e-59 and 7e6.
    variances = torch.clamp(variances, 1e-7, 1e7)

    return variances


def get_output_channels(config):
    """ Determines number of output channels based on user configuration. """
    out_channels = 0
    try:
        out_channels += len(config.data.target_surface_vars)
    except:
        print("Output does not contain surface variables")
    try:
        out_channels += (len(config.data.target_levels) * len(config.data.target_upper_level_vars))
    except:
        print("Output does not contain upper level variables")

    out_channels *= config.data.target_size_time

    return out_channels
